@@ -1,4 +1,4 @@
-// Copyright (C) 2015-2019 Internet Systems Consortium, Inc. ("ISC")
+// Copyright (C) 2015-2023 Internet Systems Consortium, Inc. ("ISC")
 //
 // This Source Code Form is subject to the terms of the Mozilla Public
 // License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -14,16 +14,17 @@
 #include <boost/multi_index/indexed_by.hpp>
 #include <boost/multi_index/member.hpp>
 #include <boost/multi_index/mem_fun.hpp>
+#include <boost/multi_index/hashed_index.hpp>
 #include <boost/multi_index/ordered_index.hpp>
 #include <boost/multi_index_container.hpp>
 #include <boost/multi_index/composite_key.hpp>
 
+#include <functional>
 #include <vector>
 
 namespace isc {
 namespace dhcp {
 
-
 /// @brief Tag for indexes by address.
 struct AddressIndexTag { };
 
@@ -33,24 +34,30 @@ struct DuidIaidTypeIndexTag { };
 /// @brief Tag for indexes by expiration time.
 struct ExpirationIndexTag { };
 
-/// @brief Tag for indexes by HW address, subnet identifier tuple.
+/// @brief Tag for indexes by HW address, subnet-id tuple.
 struct HWAddressSubnetIdIndexTag { };
 
-/// @brief Tag for indexes by client and subnet identifiers.
+/// @brief Tag for indexes by client-id, subnet-id tuple.
 struct ClientIdSubnetIdIndexTag { };
 
-/// @brief Tag for indexes by client id, HW address and subnet id.
-struct ClientIdHWAddressSubnetIdIndexTag { };
-
-/// @brief Tag for indexs by subnet-id.
+/// @brief Tag for indexes by subnet-id.
 struct SubnetIdIndexTag { };
 
+/// @brief Tag for indexes by subnet-id and pool-id.
+struct SubnetIdPoolIdIndexTag { };
+
 /// @brief Tag for index using DUID.
 struct DuidIndexTag { };
 
 /// @brief Tag for index using hostname.
 struct HostnameIndexTag { };
 
+/// @brief Tag for index using remote-id.
+struct RemoteIdIndexTag { };
+
+/// @brief Tag for index using relay-id.
+struct RelayIdIndexTag { };
+
 /// @name Multi index containers holding DHCPv4 and DHCPv6 leases.
 ///
 //@{
@@ -62,8 +69,10 @@ struct HostnameIndexTag { };
 /// - using a composite index: DUID, IAID and lease type.
 /// - using a composite index: boolean flag indicating if the state is
 ///   "expired-reclaimed" and expiration time.
+/// - using subnet ID.
+/// - using hostname.
 ///
-/// Indexes can be accessed using the index number (from 0 to 2) or a
+/// Indexes can be accessed using the index number (from 0 to 5) or a
 /// name tag. It is recommended to use the tags to access indexes as
 /// they do not depend on the order of indexes in the container.
 typedef boost::multi_index_container<
@@ -99,9 +108,9 @@ typedef boost::multi_index_container<
         // Specification of the third index starts here.
         boost::multi_index::ordered_non_unique<
             boost::multi_index::tag<ExpirationIndexTag>,
-            // This is a composite index that will be used to search for
+            // This is a composite index that is used to search for
             // the expired leases. Depending on the value of the first component
-            // of the search key, the reclaimed or not reclaimed leases will can
+            // of the search key, the reclaimed or not reclaimed leases can
             // be searched.
             boost::multi_index::composite_key<
                 Lease6,
@@ -136,6 +145,26 @@ typedef boost::multi_index_container<
         boost::multi_index::ordered_non_unique<
             boost::multi_index::tag<HostnameIndexTag>,
             boost::multi_index::member<Lease, std::string, &Lease::hostname_>
+        >,
+
+        // Specification of the seventh index starts here.
+        boost::multi_index::ordered_non_unique<
+            boost::multi_index::tag<SubnetIdPoolIdIndexTag>,
+            // This is a composite index that combines two attributes of the
+            // Lease6 object: subnet id and pool id.
+            boost::multi_index::composite_key<
+                Lease6,
+                // The subnet id is held in the subnet_id_ member of Lease6
+                // class. Note that the subnet_id_ is defined in the base
+                // class (Lease) so we have to point to this class rather
+                // than derived class: Lease6.
+                boost::multi_index::member<Lease, SubnetID, &Lease::subnet_id_>,
+                // The pool id is held in the pool_id_ member of Lease6
+                // class. Note that the pool_id_ is defined in the base
+                // class (Lease) so we have to point to this class rather
+                // than derived class: Lease6.
+                boost::multi_index::member<Lease, uint32_t, &Lease::pool_id_>
+            >
         >
     >
 > Lease6Storage; // Specify the type name of this container.
@@ -143,14 +172,17 @@ typedef boost::multi_index_container<
 /// @brief A multi index container holding DHCPv4 leases.
 ///
 /// The leases in the container may be accessed using different indexes:
-/// - IPv6 address,
-/// - composite index: HW address and subnet id,
+/// - IPv4 address,
+/// - composite index: hardware address and subnet id,
 /// - composite index: client id and subnet id,
-/// - composite index: HW address, client id and subnet id
 /// - using a composite index: boolean flag indicating if the state is
 ///   "expired-reclaimed" and expiration time.
+/// - using subnet id.
+/// - using hostname.
+/// - using remote id.
+/// - using a composite index:
 ///
-/// Indexes can be accessed using the index number (from 0 to 4) or a
+/// Indexes can be accessed using the index number (from 0 to 5) or a
 /// name tag. It is recommended to use the tags to access indexes as
 /// they do not depend on the order of indexes in the container.
 typedef boost::multi_index_container<
@@ -206,28 +238,6 @@ typedef boost::multi_index_container<
         >,
 
         // Specification of the fourth index starts here.
-        boost::multi_index::ordered_non_unique<
-            boost::multi_index::tag<ClientIdHWAddressSubnetIdIndexTag>,
-            // This is a composite index that uses three values to search for a
-            // lease: client id, HW address and subnet id.
-            boost::multi_index::composite_key<
-                Lease4,
-                // The client id can be retrieved from the Lease4 object by
-                // calling getClientIdVector const function.
-                boost::multi_index::const_mem_fun<Lease4, const std::vector<uint8_t>&,
-                                                  &Lease4::getClientIdVector>,
-                // The hardware address is held in the hwaddr_ object. We can
-                // access the raw data using lease->hwaddr_->hwaddr_, but Boost
-                // doesn't seem to provide a way to use member of a member for this,
-                // so we need a simple key extractor method (getHWAddrVector).
-                boost::multi_index::const_mem_fun<Lease, const std::vector<uint8_t>&,
-                                                  &Lease::getHWAddrVector>,
-                // The subnet id is accessed through the subnet_id_ member.
-                boost::multi_index::member<Lease, SubnetID, &Lease::subnet_id_>
-            >
-        >,
-
-        // Specification of the fifth index starts here.
         boost::multi_index::ordered_non_unique<
             boost::multi_index::tag<ExpirationIndexTag>,
             // This is a composite index that will be used to search for
@@ -245,19 +255,66 @@ typedef boost::multi_index_container<
             >
         >,
 
-        // Specification of the sixth index starts here.
+        // Specification of the fifth index starts here.
         // This index sorts leases by SubnetID.
         boost::multi_index::ordered_non_unique<
             boost::multi_index::tag<SubnetIdIndexTag>,
             boost::multi_index::member<Lease, isc::dhcp::SubnetID, &Lease::subnet_id_>
         >,
 
-
-        // Specification of the seventh index starts here
+        // Specification of the sixth index starts here.
         // This index is used to retrieve leases for matching hostname.
         boost::multi_index::ordered_non_unique<
             boost::multi_index::tag<HostnameIndexTag>,
             boost::multi_index::member<Lease, std::string, &Lease::hostname_>
+        >,
+
+        // Specification of the seventh index starts here.
+        // This index is used to retrieve leases for matching remote id
+        // for Bulk Lease Query.
+        boost::multi_index::hashed_non_unique<
+            boost::multi_index::tag<RemoteIdIndexTag>,
+            boost::multi_index::member<Lease4,
+                                       std::vector<uint8_t>,
+                                       &Lease4::remote_id_>
+        >,
+
+        // Specification of the eighth index starts here.
+        // This index is used to retrieve leases for matching relay id
+        // for Bulk Lease Query.
+        boost::multi_index::ordered_non_unique<
+            boost::multi_index::tag<RelayIdIndexTag>,
+            boost::multi_index::composite_key<
+                Lease4,
+                // Relay id.
+                boost::multi_index::member<Lease4,
+                                           std::vector<uint8_t>,
+                                           &Lease4::relay_id_>,
+                // Address.
+                boost::multi_index::member<Lease,
+                                           isc::asiolink::IOAddress,
+                                           &Lease::addr_>
+            >
+        >,
+
+        // Specification of the ninth index starts here.
+        boost::multi_index::ordered_non_unique<
+            boost::multi_index::tag<SubnetIdPoolIdIndexTag>,
+            // This is a composite index that combines two attributes of the
+            // Lease4 object: subnet id and pool id.
+            boost::multi_index::composite_key<
+                Lease4,
+                // The subnet id is held in the subnet_id_ member of Lease4
+                // class. Note that the subnet_id_ is defined in the base
+                // class (Lease) so we have to point to this class rather
+                // than derived class: Lease4.
+                boost::multi_index::member<Lease, SubnetID, &Lease::subnet_id_>,
+                // The pool id is held in the pool_id_ member of Lease4
+                // class. Note that the pool_id_ is defined in the base
+                // class (Lease) so we have to point to this class rather
+                // than derived class: Lease4.
+                boost::multi_index::member<Lease, uint32_t, &Lease::pool_id_>
+            >
         >
     >
 > Lease4Storage; // Specify the type name for this container.
@@ -277,10 +334,13 @@ typedef Lease6Storage::index<DuidIaidTypeIndexTag>::type Lease6StorageDuidIaidTy
 /// @brief DHCPv6 lease storage index by expiration time.
 typedef Lease6Storage::index<ExpirationIndexTag>::type Lease6StorageExpirationIndex;
 
-/// @brief DHCPv6 lease storage index by Subnet-id.
+/// @brief DHCPv6 lease storage index by subnet-id.
 typedef Lease6Storage::index<SubnetIdIndexTag>::type Lease6StorageSubnetIdIndex;
 
-/// @brief DHCPv6 lease storage index by Subnet-id.
+/// @brief DHCPv6 lease storage index subnet-id and pool-id.
+typedef Lease6Storage::index<SubnetIdPoolIdIndexTag>::type Lease6StorageSubnetIdPoolIdIndex;
+
+/// @brief DHCPv6 lease storage index by DUID.
 typedef Lease6Storage::index<DuidIndexTag>::type Lease6StorageDuidIndex;
 
 /// @brief DHCPv6 lease storage index by hostname.
@@ -292,25 +352,161 @@ typedef Lease4Storage::index<AddressIndexTag>::type Lease4StorageAddressIndex;
 /// @brief DHCPv4 lease storage index by expiration time.
 typedef Lease4Storage::index<ExpirationIndexTag>::type Lease4StorageExpirationIndex;
 
-/// @brief DHCPv4 lease storage index by HW address and subnet identifier.
+/// @brief DHCPv4 lease storage index by HW address and subnet-id.
 typedef Lease4Storage::index<HWAddressSubnetIdIndexTag>::type
 Lease4StorageHWAddressSubnetIdIndex;
 
-/// @brief DHCPv4 lease storage index by client and subnet identifier.
+/// @brief DHCPv4 lease storage index by client-id and subnet-id.
 typedef Lease4Storage::index<ClientIdSubnetIdIndexTag>::type
 Lease4StorageClientIdSubnetIdIndex;
 
-/// @brief DHCPv4 lease storage index by client id, HW address and subnet id.
-typedef Lease4Storage::index<ClientIdHWAddressSubnetIdIndexTag>::type
-Lease4StorageClientIdHWAddressSubnetIdIndex;
-
-/// @brief DHCPv4 lease storage index by client id, HW address and subnet id.
+/// @brief DHCPv4 lease storage index subnet-id.
 typedef Lease4Storage::index<SubnetIdIndexTag>::type Lease4StorageSubnetIdIndex;
 
+/// @brief DHCPv4 lease storage index subnet-id and pool-id.
+typedef Lease4Storage::index<SubnetIdPoolIdIndexTag>::type Lease4StorageSubnetIdPoolIdIndex;
+
 /// @brief DHCPv4 lease storage index by hostname.
 typedef Lease4Storage::index<HostnameIndexTag>::type Lease4StorageHostnameIndex;
 
+/// @brief DHCPv4 lease storage index by remote-id.
+typedef Lease4Storage::index<RemoteIdIndexTag>::type Lease4StorageRemoteIdIndex;
+
+/// @brief DHCPv4 lease storage range by remote-id.
+typedef std::pair<Lease4StorageRemoteIdIndex::const_iterator,
+                  Lease4StorageRemoteIdIndex::const_iterator> Lease4StorageRemoteIdRange;
+
+/// @brief DHCPv4 lease storage index by relay-id.
+typedef Lease4Storage::index<RelayIdIndexTag>::type Lease4StorageRelayIdIndex;
+
 //@}
+
+/// @name Multi index containers holding DHCPv6 lease extended informations
+/// for Bulk Lease Query.
+//@{
+
+/// @brief Lease6 extended informations for Bulk Lease Query.
+class Lease6ExtendedInfo {
+public:
+    /// @brief Constructor.
+    ///
+    /// @param lease_addr Lease address.
+    /// @param id Identifier.
+    Lease6ExtendedInfo(const isc::asiolink::IOAddress& lease_addr,
+                       const std::vector<uint8_t>& id)
+        : lease_addr_(lease_addr), id_(id) {
+    }
+
+    /// @brief Lease address.
+    isc::asiolink::IOAddress lease_addr_;
+
+    /// @brief Remote or relay opaque identifier.
+    std::vector<uint8_t> id_;
+};
+
+/// @brief Pointer to a Lease6ExtendedInfo object.
+typedef boost::shared_ptr<Lease6ExtendedInfo> Lease6ExtendedInfoPtr;
+
+/// @brief Tag for indexes by lease address.
+struct LeaseAddressIndexTag { };
+
+/// @brief A multi index container holding lease6 extended info for by relay id.
+///
+/// The lease6 extended info may be accessed using different indexes:
+/// - using relay id, and lease address for getting lower bounds.
+/// - using lease address for deletes.
+///
+/// The choice of binary trees was governed by the fact a large number of
+/// clients can be behind a relay.
+///
+/// Indexes can be accessed using the index number (from 0 to 5) or a
+/// name tag. It is recommended to use the tags to access indexes as
+/// they do not depend on the order of indexes in the container.
+typedef boost::multi_index_container<
+    // It holds pointers to lease6 extended info.
+    Lease6ExtendedInfoPtr,
+    boost::multi_index::indexed_by<
+        // First index is by relay id and lease address.
+        boost::multi_index::ordered_non_unique<
+            boost::multi_index::tag<RelayIdIndexTag>,
+            boost::multi_index::composite_key<
+                Lease6ExtendedInfo,
+                boost::multi_index::member<Lease6ExtendedInfo,
+                                           std::vector<uint8_t>,
+                                           &Lease6ExtendedInfo::id_>,
+                boost::multi_index::member<Lease6ExtendedInfo,
+                                           isc::asiolink::IOAddress,
+                                           &Lease6ExtendedInfo::lease_addr_>
+            >
+        >,
+
+        // Last index is by lease address.
+        boost::multi_index::hashed_non_unique<
+            boost::multi_index::tag<LeaseAddressIndexTag>,
+            boost::multi_index::member<Lease6ExtendedInfo,
+                                       isc::asiolink::IOAddress,
+                                       &Lease6ExtendedInfo::lease_addr_>
+        >
+    >
+> Lease6ExtendedInfoRelayIdTable;
+
+/// @brief A multi index container holding lease6 extended info for by remote id.
+///
+/// The lease6 extended info may be accessed using different indexes:
+/// - using remote id.
+/// - using lease address for deletes.
+///
+/// The internal layout of remote id is not used. The choice of hash tables
+/// was governed by the fact a small number of clients should share the same
+/// remote id.
+///
+/// Indexes can be accessed using the index number (from 0 to 5) or a
+/// name tag. It is recommended to use the tags to access indexes as
+/// they do not depend on the order of indexes in the container.
+typedef boost::multi_index_container<
+    // It holds pointers to lease6 extended info.
+    Lease6ExtendedInfoPtr,
+    boost::multi_index::indexed_by<
+        // First index is by remote id.
+        boost::multi_index::hashed_non_unique<
+            boost::multi_index::tag<RemoteIdIndexTag>,
+            boost::multi_index::member<Lease6ExtendedInfo,
+                                       std::vector<uint8_t>,
+                                       &Lease6ExtendedInfo::id_>
+        >,
+
+        // Last index is by lease address.
+        boost::multi_index::hashed_non_unique<
+            boost::multi_index::tag<LeaseAddressIndexTag>,
+            boost::multi_index::member<Lease6ExtendedInfo,
+                                       isc::asiolink::IOAddress,
+                                       &Lease6ExtendedInfo::lease_addr_>
+        >
+    >
+> Lease6ExtendedInfoRemoteIdTable;
+
+/// @brief Lease6 extended information by relay id index.
+typedef Lease6ExtendedInfoRelayIdTable::index<RelayIdIndexTag>::type
+    RelayIdIndex;
+
+/// @brief Lease6 extended information by lease address index of by relay id table.
+typedef Lease6ExtendedInfoRelayIdTable::index<LeaseAddressIndexTag>::type
+    LeaseAddressRelayIdIndex;
+
+/// @brief Lease6 extended information by remote id index.
+typedef Lease6ExtendedInfoRemoteIdTable::index<RemoteIdIndexTag>::type
+    RemoteIdIndex;
+
+/// @brief Lease6 extended information by remote id range.
+typedef std::pair<RemoteIdIndex::const_iterator, RemoteIdIndex::const_iterator>
+    RemoteIdIndexRange;
+
+/// @brief Lease6 extended information by lease address index of by remote id table.
+typedef Lease6ExtendedInfoRemoteIdTable::index<LeaseAddressIndexTag>::type
+    LeaseAddressRemoteIdIndex;
+
+//@}
+
 } // end of isc::dhcp namespace
 } // end of isc namespace
 
